/*  Copyright (C) 2023 Tibor 'Igor2' Palinkas - license: Creative Commons CC0 */

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>

#include <libgpcogl/libgpcogl.h>
#include <libgpcogl/gli.h>

#include "input.c"

#define NL "\n"

/* Read current state from src_state and boundary conditions from input_temp
   and combine them so that output has input_temp where input_temp is not
   zero and src_state elsewhere. Typically src_state and output are the same
   array and the boundary conditions from input_temp are masked onto that
   array. */
const char *pre_iterate_source =
	NL GPCOGL_SHADER_PROG(130)                          /* program header, require glsl ver 1.3 */
	NL GPCOGL_OUT_1_FLOAT(output_cell)                  /* specify output cell type (1*float), create a variable output_cell for the output */
	NL GPCOGL_IN_1_FLOAT(src_state)                     /* specify an input array with cells (1*float); this is where the current state is read from */
	NL GPCOGL_IN_1_FLOAT(input_temp)                    /* specify an input array with cells (1*float); this is where input temperatures are read from */
	NL "void main()"                                    /* shader program entry point */
	NL "{"
	NL "	" GPCOGL_OUT_COORDS(x, y)                     /* fetch output cell coords into variables x and y */
	NL "	float in_t = getxy_input_temp(x, y);"         /* read input_temp cell in a temp var so the computation is easier to read */
	NL "	float src_t = getxy_src_state(x, y);"         /* read src_state cell in a temp var so the computation is easier to read */
	NL "	output_cell = (in_t > 0) ? in_t : src_t;"     /* output is input_temp at this point, if it's not 0; else copy the original source cell at this point */
	NL "}";

/* Calculate new state from prev state. */
const char *iterate_source =
	NL GPCOGL_SHADER_PROG(130)                          /* program header, require glsl ver 1.3 */
	NL GPCOGL_OUT_1_FLOAT(output_cell)                  /* specify output cell type (1*float), create a variable output_cell for the output */
	NL GPCOGL_IN_1_FLOAT(prev)                          /* specify an input array with cells (1*float); this is where the previous state is read from */
	NL "void main()"                                    /* shader program entry point */
	NL "{"
	NL "	" GPCOGL_OUT_COORDS(x, y)                     /* fetch output cell coords into variables x and y */
	NL "	float left = getxy_prev(x-1, y);"             /* read previous state neighbours into temp vars... */
	NL "	float right = getxy_prev(x+1, y);"
	NL "	float up = getxy_prev(x, y-1);"
	NL "	float down = getxy_prev(x, y+1);"
	NL "	output_cell = (left + right + up + down) / 4.0;" /* current state is the average of the 4 neighbors from the previous state */
	NL "}";

int main(int argc, char **argv)
{
	gpcogl_prog_id_t prg_pre_it;  /* shader prg handle: combine state and initial state */
	gpcogl_prog_id_t prg_it;      /* shader prg handle: calculate new state from previous state */
	gpcogl_t ctx;                 /* lib context all arrays and shader programs are in */
	gpcogl_array_id_t ainp;       /* GPU-handle of the static input temp array */
	gpcogl_array_id_t ast1, ast2; /* GPU-handle of the temporary state arrays */
	float *res;                   /* CPU-side copy of the output array */
	int n;

	/* Initialize the opengl interface library */
	gpcogl_gli_init(&argc, argv);

	/* Create a context (hidden X window with glut) for off-screen rendering */
	if (gpcogl_gli_make_context(&ctx, NULL) != 0)
		return 1;

	/* Compile the shader programs */
	prg_pre_it = gpcogl_compile(&ctx, pre_iterate_source);
	if (prg_pre_it == 0) {
		fprintf(stderr, "Failed to compile prg_pre_it\n");
		return 1;
	}

	prg_it = gpcogl_compile(&ctx, iterate_source);
	if (prg_it == 0) {
		fprintf(stderr, "Failed to compile prg_it\n");
		return 1;
	}


	/* create empty state arrays in the GPU */
	ast1 = gpcogl_array_create(&ctx, GPCOGL_CT_1_FLOAT32, SX, SY);
	ast2 = gpcogl_array_create(&ctx, GPCOGL_CT_1_FLOAT32, SX, SY);

	/* reset state arrays to all-zero; cells where input_temps is 0 will not
	   be initialized otherwise */
	gpcogl_array_clear(&ctx, ast1);
	gpcogl_array_clear(&ctx, ast2);

	/* upload the initial state array */
	ainp = gpcogl_array_init(&ctx, GPCOGL_CT_1_FLOAT32, input_temps, SX, SY);

	for(n = 0; n < 100; n++) {
		gpcogl_vcompute(&ctx, prg_pre_it, ast1,   "src_state", ast1, "input_temp", ainp, NULL);
		gpcogl_vcompute(&ctx, prg_it,     ast2,   "prev", ast1, NULL);
		gpcogl_vcompute(&ctx, prg_pre_it, ast2,   "src_state", ast2, "input_temp", ainp, NULL);
		gpcogl_vcompute(&ctx, prg_it,     ast1,   "prev", ast2, NULL);
	}

	/* Allocate memory in CPU ram for the result and download from "ot" to res */
	res = calloc(SX*SY, sizeof(float));
	gpcogl_vcompute(&ctx, prg_pre_it, ast1,   "src_state", ast1, "input_temp", ainp, NULL);
	gpcogl_array_get(&ctx, res, ast1);

	/* print result */
	print_mx(res);

	/* free context and lib globals */
	gpcogl_gli_destroy_context(&ctx);
	gpcogl_gli_uninit();

	return 0;
}
