/*
    libgpcogl - Generic Purpose Computation on OpenGL (using shaders)
    Copyright (C) 2023 Tibor 'Igor2' Palinkas

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA

    Contact:
      Project page: http://www.repo.hu/projects/libgpcogl
      Lead developer: http://www.repo.hu/projects/libgpcogl/contact.html
*/


#ifndef LIBGPCOGL_H
#define LIBGPCOGL_H

#include <GL/gl.h>

/* API version of the library: major.minor.patch */
extern const int gpcogl_ver_major, gpcogl_ver_minor, gpcogl_ver_patch;

/* Each (shader) program compiled is remembered by a gpcogl_prog_id_t
   (per gpcogl_t context) assigned by the compiler */
typedef GLuint gpcogl_prog_id_t;

/* Each array ever created in a gpcogl_t context gets an unique ID of
   gpcogl_array_id_t; this is used to pass it to compute() calls and to
   free the array later */
typedef struct gpcogl_array_id_s gpcogl_array_id_t;

typedef struct gpcogl_s gpcogl_t;

struct gpcogl_s {
	/*** public: GPU capabilities ***/

	/* maximum size of any array in x or y direction */
	GLint array_max_size;

	/* maximum number of arrays that can be accessed by a program (there can
	   be more arrays uploaded in GPU memory, but they are passive,
	   unaccessible to the current program). At the end a program can have
	   one output array,  and (array_max_active-2) input arrays.
	   Guarantee: array_max_active >= 16. */
	GLint array_max_active;

	/* Vendor and version */
	GLint glver_major, glver_minor;
	const unsigned char *gl_vendor;
	const unsigned char *renderer_vendor;
	

	/*** private: internals ***/
	int next_texture_unit;

	/* Make this ctx's context current; implemented in the gl interface lib */
	int (*make_current)(gpcogl_t *ctx);

	/* Render onto the screen (into the currently active window); display is
	   called whenever rendering is needed, with the window context activated */
	void (*screen_render)(gpcogl_t *ctx, void (*display)(void *user_data), void *user_data);

	/* If not NULL, use this instead of glFlush() for flushing the output
	   in a gl interface lib specific way */
	void (*flush)(gpcogl_t *ctx);

	struct {
		long l;
		void *p0, *p1, *p2;
	} ogl_ctx; /* whatever the gl interface lib (e.g. freeglut) needs to remember the context */
};

typedef enum gpcogl_cell_type_e {   /* each cell of the array is holding ... */
	GPCOGL_CT_1_FLOAT32,              /* ... a single float32 */
	GPCOGL_CT_2_FLOAT32,              /* ... a pair of float32s */
	GPCOGL_CT_3_FLOAT32,              /* ... three of float32s */
	GPCOGL_CT_4_FLOAT32,              /* ... four of float32s */
	GPCOGL_CT_1_INT32,                /* ... a single int32 */
	GPCOGL_CT_2_INT32,                /* ... a pair of int32s */
	GPCOGL_CT_3_INT32,                /* ... three of int32s */
	GPCOGL_CT_4_INT32,                /* ... four of int32s */
	GPCOGL_CT_1_UINT32,               /* ... a single uint32 */
	GPCOGL_CT_2_UINT32,               /* ... a pair of uint32s */
	GPCOGL_CT_3_UINT32,               /* ... three of uint32s */
	GPCOGL_CT_4_UINT32,               /* ... four of uint32s */

	GPCOGL_CT_max
} gpcogl_cell_type_t;

struct gpcogl_array_id_s {
	GLuint tid;       /* texture ID */
	unsigned ct:4;    /* of gpcogl_cell_type_t */
};

/* Invalid array ID - returned as error */
extern const gpcogl_array_id_t gpcogl_no_array;


/* Compile a fragment shader program source; the returned program ID can
   be used in gpcogl_*compute() to do computations on arrays. Returns 0 on
   error. */
gpcogl_prog_id_t gpcogl_compile(gpcogl_t *ctx, const char *fragment_shader);

/* Execute a program to perform computations. Output is written to the
   dst array. Input is a list of name-array_id pairs, where name must
   match the name of the input in the shader prorgam. There can be no
   more input arrays than ctx->array_max_active-2, but at least
   14 input arrays must be supported by the hardware.

   In the 'a' variant input is passed in the src_name and src_aid
   arrays, where src_name is NULL terminated. Arguments const_type, const_name,
   cont_val behave the same, for global constants; const_type is a NULL
   terminated array, each element is one of GPCOGL_C_*. For each non-NULL
   type there is a corresponding array element in const_name and const_val.
   These constants are made acessible within the shader program
   using the GPCOGL_CONST_* macros.

   The 'v' variant takes a NULL terminated vararg list of mixed array and
   global constant definitions. An array definition is of form
   name,array_id pair. A global constant is type,name,val, where
   type is one of GPCOGL_C_*, name is a (const char *) string and
   val is a (float *) or (int32_t *) or (uint32_t *) depending on type.

   The _box variants operate on a rectangular subset of the
   output. The target rectangle's first cell affected (top left) is x0;y0,
   cells being indexed from 0. Target rectangle width and height are w and
   h, in number-of-cells. If w or h is 0, it is calculated from array
   width or height so that it spans the whole remaining array in that direction.
   Thus a 0,0,0,0 means "whole array", while a 0,3,0,2 means "two whole rows
   of the array starting from row 3". */
int gpcogl_acompute(gpcogl_t *ctx, gpcogl_prog_id_t pid, gpcogl_array_id_t dst, const char **src_name, gpcogl_array_id_t *src_aid, const char **const_type, const char **const_name, const void **const_val);
int gpcogl_acompute_box(gpcogl_t *ctx, gpcogl_prog_id_t pid, gpcogl_array_id_t dst, GLint x0, GLint y0, GLsizei w, GLsizei h, const char **src_name, gpcogl_array_id_t *src_aid, const char **const_type, const char **const_name, const void **const_val);
int gpcogl_vcompute(gpcogl_t *ctx, gpcogl_prog_id_t pid, gpcogl_array_id_t dst, ...);
int gpcogl_vcompute_box(gpcogl_t *ctx, gpcogl_prog_id_t pid, gpcogl_array_id_t dst, GLint x0, GLint y0, GLsizei w, GLsizei h, ...);

/* const prefixes for gpcogl_vcompute*() calls */
extern const char GPCOGL_C_1F[], GPCOGL_C_2F[], GPCOGL_C_3F[], GPCOGL_C_4F[];
extern const char GPCOGL_C_1U[], GPCOGL_C_2U[], GPCOGL_C_3U[], GPCOGL_C_4U[];
extern const char GPCOGL_C_1I[], GPCOGL_C_2I[], GPCOGL_C_3I[], GPCOGL_C_4I[];


/* Allocate an uninitialized array in the GPU, return the ID of the array. */
gpcogl_array_id_t gpcogl_array_create(gpcogl_t *ctx, gpcogl_cell_type_t ct, int sx, int sy);

/* Set all components of all cells of dst to 0. Returns 0 on success. */
int gpcogl_array_clear(gpcogl_t *ctx, gpcogl_array_id_t dst);

/* Upload an array from src in CPU to dst in the GPU; same as
   gpcogl_array_create() except that the data in the array is initialized
   from inp. */
gpcogl_array_id_t gpcogl_array_init(gpcogl_t *ctx, gpcogl_cell_type_t ct, const void *inp, int sx, int sy);

/* Get an array from GPU to dst; dst shall be large enough to hold sx*sy cells,
   where sx and sy are array size. Returns 0 on success. */
int gpcogl_array_get(gpcogl_t *ctx, void *dst, gpcogl_array_id_t src);

/* Same as gpcogl_array_get(), but performs a partial read of w*h cells */
int gpcogl_array_get_box(gpcogl_t *ctx, void *dst, gpcogl_array_id_t src, GLint x0, GLint y0, GLsizei w, GLsizei h);

/* Fetches size of an array into sx and sy; returns 0 on success */
int gpcogl_array_get_size(gpcogl_t *ctx, GLint *sx, GLint *sy, gpcogl_array_id_t src);

/* Remove an array from GPU memory. Returns 0 on success. */
int gpcogl_array_free(gpcogl_t *ctx, gpcogl_array_id_t aid);

/* Blocking-wait until all current computations are finished in ctx */
void gpcogl_finish(gpcogl_t *ctx);

/*** shader macros ***/

/* Internal macro to switch over line numbering to C source line numbering */
#define GPCOGL_TO_STR_(x) #x
#define GPCOGL_TO_STR(x) GPCOGL_TO_STR_(x)
#define GPCOGL_LINE_FIX "\n#line " GPCOGL_TO_STR(__LINE__)

#define GPCOGL_SHADER_PROG(minver) \
	"#version " GPCOGL_TO_STR(minver) "\n" \
	"#extension GL_ARB_texture_rectangle : enable\n"  \
	GPCOGL_LINE_FIX


/* Declare int variables xname and yname and load them with the x and y
   coord of the current output array cell. Both xname and yname are shader
   variable names to be created. (Using signed integers for x and y makes it
   easier to do things like +1 or -1 without having to convert.) */
#define GPCOGL_OUT_COORDS(xname, yname) \
	"	int " #xname " = int(gl_FragCoord.x);" \
	"	int " #yname " = int(gl_FragCoord.y);" \
	GPCOGL_LINE_FIX


/* Declare input array by cell type and name. Name is a valid identifier.
   This is the same name used as src_name in gpcogl_*compute(). Each name is
   declared as a shader sampler plus a getxy_name(x,y) accessor is created that
   returns one element of the input array addressed by float x,y (truncated to
   ints). Return type matches cell type: float, int or uint for IN_1_ and
   vec[23], ivec[23] or uvec[23] for IN_[23]_*. */
#define GPCOGL_IN_1_FLOAT(name)  "uniform sampler2DRect " #name ";\n#define getxy_" #name "(x,y) texture(" #name ", vec2(x, y)).r" GPCOGL_LINE_FIX
#define GPCOGL_IN_2_FLOAT(name)  "uniform sampler2DRect " #name ";\n#define getxy_" #name "(x,y) vec2(texture(" #name ", vec2(x, y)))" GPCOGL_LINE_FIX
#define GPCOGL_IN_3_FLOAT(name)  "uniform sampler2DRect " #name ";\n#define getxy_" #name "(x,y) vec3(texture(" #name ", vec2(x, y)))" GPCOGL_LINE_FIX
#define GPCOGL_IN_4_FLOAT(name)  "uniform sampler2DRect " #name ";\n#define getxy_" #name "(x,y) texture(" #name ", vec2(x, y))" GPCOGL_LINE_FIX
#define GPCOGL_IN_1_INT32(name)  "uniform sampler2DRect " #name ";\n#define getxy_" #name "(x,y) int(texture(" #name ", vec2(x, y)).r)" GPCOGL_LINE_FIX
#define GPCOGL_IN_2_INT32(name)  "uniform sampler2DRect " #name ";\n#define getxy_" #name "(x,y) ivec2(texture(" #name ", vec2(x, y)))" GPCOGL_LINE_FIX
#define GPCOGL_IN_3_INT32(name)  "uniform sampler2DRect " #name ";\n#define getxy_" #name "(x,y) ivec3(texture(" #name ", vec2(x, y)))" GPCOGL_LINE_FIX
#define GPCOGL_IN_4_INT32(name)  "uniform sampler2DRect " #name ";\n#define getxy_" #name "(x,y) ivec4(texture(" #name ", vec2(x, y)))" GPCOGL_LINE_FIX
#define GPCOGL_IN_1_UINT32(name)  "uniform sampler2DRect " #name ";\n#define getxy_" #name "(x,y) uint(texture(" #name ", vec2(x, y)).r)" GPCOGL_LINE_FIX
#define GPCOGL_IN_2_UINT32(name)  "uniform sampler2DRect " #name ";\n#define getxy_" #name "(x,y) uvec2(texture(" #name ", vec2(x, y)))" GPCOGL_LINE_FIX
#define GPCOGL_IN_3_UINT32(name)  "uniform sampler2DRect " #name ";\n#define getxy_" #name "(x,y) uvec3(texture(" #name ", vec2(x, y)))" GPCOGL_LINE_FIX
#define GPCOGL_IN_4_UINT32(name)  "uniform sampler2DRect " #name ";\n#define getxy_" #name "(x,y) uvec4(texture(" #name ", vec2(x, y)))" GPCOGL_LINE_FIX

/* Declare output array cell variable by cell type and name. Name is the
   name of a shader variable to be created, e.g. "out". After this, the
   shader program can assign value to the named variable to produce its
   output, e.g. out = 42. */
#define GPCOGL_OUT_1_FLOAT(name)  "out float " #name ";" GPCOGL_LINE_FIX
#define GPCOGL_OUT_2_FLOAT(name)  "out vec2 " #name ";" GPCOGL_LINE_FIX
#define GPCOGL_OUT_3_FLOAT(name)  "out vec3 " #name ";" GPCOGL_LINE_FIX
#define GPCOGL_OUT_4_FLOAT(name)  "out vec4 " #name ";" GPCOGL_LINE_FIX
#define GPCOGL_OUT_1_INT32(name)  "out int " #name ";" GPCOGL_LINE_FIX
#define GPCOGL_OUT_2_INT32(name)  "out ivec2 " #name ";" GPCOGL_LINE_FIX
#define GPCOGL_OUT_3_INT32(name)  "out ivec3 " #name ";" GPCOGL_LINE_FIX
#define GPCOGL_OUT_4_INT32(name)  "out ivec4 " #name ";" GPCOGL_LINE_FIX
#define GPCOGL_OUT_1_UINT32(name) "out uint " #name ";" GPCOGL_LINE_FIX
#define GPCOGL_OUT_2_UINT32(name) "out uvec2 " #name ";" GPCOGL_LINE_FIX
#define GPCOGL_OUT_3_UINT32(name) "out uvec3 " #name ";" GPCOGL_LINE_FIX
#define GPCOGL_OUT_4_UINT32(name) "out uvec4 " #name ";" GPCOGL_LINE_FIX

/* Declare a global constant by type and name. The global const can be
   used as a read-only variable by the shader code. */
#define GPCOGL_CONST_1_FLOAT(name) "uniform float " #name ";" GPCOGL_LINE_FIX
#define GPCOGL_CONST_2_FLOAT(name) "uniform vec2 " #name ";" GPCOGL_LINE_FIX
#define GPCOGL_CONST_3_FLOAT(name) "uniform vec3 " #name ";" GPCOGL_LINE_FIX
#define GPCOGL_CONST_4_FLOAT(name) "uniform vec4 " #name ";" GPCOGL_LINE_FIX
#define GPCOGL_CONST_1_INT32(name) "uniform int " #name ";" GPCOGL_LINE_FIX
#define GPCOGL_CONST_2_INT32(name) "uniform ivec2 " #name ";" GPCOGL_LINE_FIX
#define GPCOGL_CONST_3_INT32(name) "uniform ivec3 " #name ";" GPCOGL_LINE_FIX
#define GPCOGL_CONST_4_INT32(name) "uniform ivec4 " #name ";" GPCOGL_LINE_FIX
#define GPCOGL_CONST_1_UINT32(name) "uniform uint " #name ";" GPCOGL_LINE_FIX
#define GPCOGL_CONST_2_UINT32(name) "uniform uvec2 " #name ";" GPCOGL_LINE_FIX
#define GPCOGL_CONST_3_UINT32(name) "uniform uvec3 " #name ";" GPCOGL_LINE_FIX
#define GPCOGL_CONST_4_UINT32(name) "uniform uvec4 " #name ";" GPCOGL_LINE_FIX

/*** odd use cases, e.g. screen rendering for debug ***/

/* Flush/swap - make sure screen render appears */
void gpcogl_flush(gpcogl_t *ctx);



/*** internal calls ***/

/* The interface lib calls this to apply default configuration and get
   some info fields of ctx filled in; returns 0 on success */
int gpcogl_init_context(gpcogl_t *ctx);

/* The interface lib calls this before destroying an opengl cotnext */
void gpcogl_uninit_context(gpcogl_t *ctx);

/* The interface lib calls this after creating a context to redirect the
   output from screen to frame buffer */
GLuint gpcogl_redir_render_to_fb(gpcogl_t *ctx);

#endif
